/*
 * cbus_bitbang.c:
 *
 * Bitbang driver for the CBUS 0..3 bits on Sparkfun FTDI USB/Serial breakout board
 *
 * This has been enhanced to implement a 3-wire serial interface
 * for the AR1010-based Sparkfun FM radio tuner.
 *
 * This program just implements basic register read/write functionality.
 * Everything else is left to a higher-level program/script.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <errno.h>
#include <time.h>
#include <string.h>

#include <ftdi.h>

static struct ftdi_context ftdic;
static int verbose = 0;

enum {
	CBUS0		= (1 << 0),
	CBUS1		= (1 << 1),
	CBUS2		= (1 << 2),
	CBUS3		= (1 << 3),

	rDATA		= CBUS0,
	rCLOCK		= CBUS1,
	rBUSEN		= CBUS2,
	rLED		= CBUS3,

	rMAXREG		= 0x1c,

	REG_READ	= 1,
	REG_WRITE	= 0
};

static void print_cbus_bits (const char *prefix, unsigned char bits)
{
	fprintf(stderr, "%s 0x%02x [%c %c %c %c]\n", prefix, bits,
		(bits & CBUS3) ? '3' : '-',
		(bits & CBUS2) ? '2' : '-',
		(bits & CBUS1) ? '1' : '-',
		(bits & CBUS0) ? '0' : '-');
}

/*
 * To simplify the code, we record errors from get_pins()/set_pins() here,
 * rather than returning/checking them after every single bit operation.
 */
static int get_set_pin_err;

static void get_pins (unsigned char *pins, unsigned char mask)
{
	int rc;

	if (get_set_pin_err)
		return;
	rc = ftdi_read_pins(&ftdic, pins);
	if (rc < 0) {
		fprintf(stderr, "read_pins: error %d (%s)\n", rc, ftdi_get_error_string(&ftdic));
		get_set_pin_err = EIO;
	} else {
		*pins &= mask;
		if (verbose > 1)
			print_cbus_bits("read    ", *pins);
	}
}

static void set_pins (unsigned char pins, unsigned char mask)
{
	int rc;
	unsigned char out;

	if (get_set_pin_err)
		return;
	mask &= 0x0f;
	pins &= mask;
	out   = pins | (mask << 4);
	if (verbose > 1)
		print_cbus_bits("writing ", pins);
	rc = ftdi_set_bitmode(&ftdic, out, BITMODE_CBUS);
	if (rc < 0) {
		fprintf(stderr, "set_bitmode(0x%x): error %d (%s)\n", out, rc, ftdi_get_error_string(&ftdic));
		get_set_pin_err = EIO;
	}
}

static unsigned int get_bit (void)
{
	unsigned char rdata = 0;

	set_pins(rCLOCK, rBUSEN|rCLOCK);
	get_pins(&rdata, rDATA);
	set_pins(0,      rBUSEN|rCLOCK);
	return rdata ? 1 : 0;
}

static void get_bits (unsigned int *data_p, int bit, int end)
{
	unsigned int data = 0;

	while (bit >= end)
		data |= get_bit() << bit--;
	*data_p = data;
}

static void put_bit (unsigned int data)
{
	unsigned char rdata = (data & 1) ? rDATA : 0;

	set_pins(rCLOCK|rdata,	rBUSEN|rCLOCK|rDATA);
	set_pins(       rdata,	rBUSEN|rCLOCK|rDATA);
}

static void put_bits (unsigned int data, int bit, int end)
{
	while (bit >= end)
		put_bit(data >> bit--);
}

static int radio_reg_rw (int rw, unsigned int addr, unsigned int *data_p)
{
	get_set_pin_err = 0;
	set_pins(rBUSEN|rCLOCK,	rBUSEN|rCLOCK);
	set_pins(0,		rBUSEN|rCLOCK);
	put_bits(addr, 7, 4);
	put_bit(rw);
	put_bits(addr, 3, 0);
	if (rw == REG_READ)
		get_bits(data_p, 15, 0);
	else
		put_bits(*data_p, 15, 0);
	set_pins(rBUSEN|rCLOCK,	rBUSEN|rCLOCK);
	return get_set_pin_err;
}

struct regnames_s {
	const char *name;
	int addr;
};

static struct regnames_s regnames[] = {
	{"r00",		0x00},
	{"r01",		0x01},
	{"r02",		0x02},
	{"r03",		0x03},
	{"r04",		0x04},
	{"r05",		0x05},
	{"r06",		0x06},
	{"r07",		0x07},
	{"r08",		0x08},
	{"r09",		0x09},
	{"r10",		0x0a},
	{"r11",		0x0b},
	{"r12",		0x0c},
	{"r13",		0x0d},
	{"r14",		0x0e},
	{"r15",		0x0f},
	{"r16",		0x10},
	{"r17",		0x11},
	{"rssi",	0x12},
	{"status",	0x13},
	{"rbs",		0x14},
	{"rds1",	0x15},
	{"rds2",	0x16},
	{"rds3",	0x17},
	{"rds4",	0x18},
	{"rds5",	0x19},
	{"rds6",	0x1a},
	{"devid",	0x1b},
	{"chipid",	0x1c},
	{NULL,		-1}
};

static int parse_addr (char *arg, unsigned int *addr_p)
{
	long addr;
	char *endp = NULL;
	int r;

	for (r = 0; regnames[r].name != NULL; ++r) {
		struct regnames_s *rn = &regnames[r];
		if (0 == strcmp(rn->name, arg)) {
			*addr_p = rn->addr;
			return 0;
		}
	}
	errno = 0;
	addr = strtol(arg, &endp, 0);
	if (errno || (endp && *endp) || addr < 0 || addr > rMAXREG) {
		fprintf(stderr, "%s: %s: unknown register, aborted.\n", __func__, arg);
		return EINVAL;
	}
	*addr_p = addr;
	return 0;
}

static int readall_regs (void)
{
	int r, err;

	for (r = 0; regnames[r].name != NULL; ++r) {
		struct regnames_s *rn = &regnames[r];
		unsigned int data, addr = rn->addr;
		err = radio_reg_rw(REG_READ, addr, &data);
		if (err)
			return err;
		printf("0x%02x -> 0x%04x\n", addr, data);
	}
	return 0;
}

static int do_args (int argc, char **argv)
{
	int err = 0, led = 1;

	if (argc > 0)
		--argc, ++argv;
	while (!err && argc > 0) {
		unsigned int do_read = 0, do_write = 0, addr = 0;
		const char *arg;

		arg = *argv++;
		--argc;
		if (0 == strcmp(arg, "--verbose") || 0 == strcmp(arg, "-v")) {
			verbose += 1;
			continue;
		} else if (0 == strcmp(arg, "--read") || 0 == strcmp(arg, "-r")) {
			do_read = 1;
			if (!argc) {
				fprintf(stderr, "%s: missing addr\n", arg);
				err = EINVAL;
			}
		} else if (0 == strcmp(arg, "--readall") || 0 == strcmp(arg, "-a")) {
			err = readall_regs();
			if (!err)
				continue;
		} else if (0 == strcmp(arg, "--write") || 0 == strcmp(arg, "-w")) {
			do_write = 1;
			if (!argc) {
				fprintf(stderr, "%s: missing addr\n", arg);
				err = EINVAL;
			}
			if (argc < 2) {
				fprintf(stderr, "%s: missing value\n", arg);
				err = EINVAL;
			}
		} else {
			fprintf(stderr, "%s: bad/unknown arg\n", arg);
			err = EINVAL;
		}

		if (!err) {
			--argc;
			err = parse_addr(*argv++, &addr);
		}

		if (!err && do_write) {
			char *endp = NULL;
			long val = 0;
			errno = 0;
			val = strtol(*argv++, &endp, 0);
			--argc;
			if (errno || (endp && *endp) || val < 0 || val > 0xffff) {
				fprintf(stderr, "%s: bad data value, aborted.\n", arg);
				err = EINVAL;
			} else {
				unsigned int data = val;
				err = radio_reg_rw(REG_WRITE, addr, &data);
				if (!err) {
					led = addr != 0 || data & 1;
					if (verbose)
						fprintf(stderr, "REG:0x%02x <-- 0x%04x\n", addr, (unsigned int)data);
				}
			}
		} else if (!err && do_read) {
			unsigned int data;
			err = radio_reg_rw(REG_READ, addr, &data);
			if (!err) {
				if (verbose)
					fprintf(stderr, "REG:0x%02x --> 0x%04x\n", addr, data);
				printf("0x%04x\n", data);
			}
		}
	}
	if (!err && led)
		set_pins(rBUSEN|rCLOCK|rLED, rBUSEN|rCLOCK|rLED); /* turn on (optional) LED */
	return err;
}

static const char *myname;

int main(int argc, char **argv)
{
	static const char usb_desc[] = "FM Radio Module";
	const char *tail;
	int rc, err;

	myname = argv[0];

	tail = strrchr(myname, '/');
	if (tail && *++tail)
		myname = tail;

	if (argc < 2) {
		fprintf(stderr, "Usage:  %s [--verbose|-v] [--read|-r | --write|-w] [addr] [<data>] ..\n", myname);
		exit(EINVAL);
	}

	ftdi_init(&ftdic);
	rc = ftdi_usb_open_desc(&ftdic, 0x0403, 0x6001, usb_desc, NULL);
	if (rc < 0) {
		fprintf(stderr, "Unable to open FTDI \"%s\" device: %d (%s)\n",
			usb_desc, rc, ftdi_get_error_string(&ftdic));
		err = EIO;
	} else {
		err = do_args(argc, argv);
		ftdi_usb_close(&ftdic);
	}
	ftdi_deinit(&ftdic);
	return err;
}
