#!/bin/bash
#
# Script to control AR1010-based Sparkfun FM radio module.
# This calls down to fmradio_bitbang.c to read/write the chip registers.
#
# Usage:	fmradio_control.sh 91.5
#		fmradio_control.sh cbc1
#		fmradio_control.sh --init 915
#		fmradio_control.sh --vol 15  ## on a scale from 0 to 18
#		fmradio_control.sh --off
#
# Preceed all args with one or more -v flags for debugging output.
#

#export PATH=".:$PATH"

function wait_for_tune_complete(){
	while [ $((`$FMRADIO -r status` & 0x20)) -eq 0 ]; do
		echo -n
	done
}

function mute(){
	$FMRADIO -w r01 $((R01 | muted))
}

function report_status(){
	r01=`$FMRADIO -r r01`
	if [ $((r01 & mono)) -eq 0 ]; then
		mode=stereo
	else
		mode=mono
	fi
	rssi=$((`$FMRADIO -r rssi` >> 9))
	rssi=$((rssi * 100 / 127))
	echo "${mode} ${rssi}%"
}

function unmute(){
	status="`$FMRADIO -r status`"
	r01=$((R01 & ~muted))
	[ $((status & 8)) -eq 0 ] && r01=$((r01 | mono))
	$FMRADIO -w r01 $r01
}

function set_volume(){
	step=$(($1 + 1))
	[ $step -gt 19 ] && step=19
	set 0 c d f c d f f f f f d e f e f f f f 
	v2=0x${!step}
	set f f f f b b b a 9 8 7 6 6 6 3 3 2 1 0 
	v1=0x${!step}
	r03=$(((R03 & ~0x0780) | (v1 <<  7)))
	r14=$(((R14 & ~0xf000) | (v2 << 12)))
	$FMRADIO -w r03 $r03 -w r14 $r14
}

function measure_rssi(){
	freq=$1
	r11=$2
	$FMRADIO -w r02 $((freq | tune_off)) -w r11 $r11 -w r02 $((freq | tune_on))
	wait_for_tune_complete
	rssi=$((`$FMRADIO -r rssi -w r02 $((freq | tune_off))` >> 9))
	echo "$rssi"
}

function autotune(){
	freq=$(($1 - 690))
	mute

	## Measure signal strength using loside, and again using hiside:
	rssi_loside=`measure_rssi $freq $((R11 | sample_loside))`
	rssi_hiside=`measure_rssi $freq $((R11 | sample_hiside))`

	## Pick the stronger of the two, and lock it in:
	if [ $rssi_hiside -gt $rssi_loside ]; then
		r11=$((R11 | hiside))
	else
		r11=$((R11 | loside))
	fi
	$FMRADIO -w r11 $r11 -w r02 $((freq | tune_on))
	wait_for_tune_complete
	unmute
}

export R01=0x5b15
export muted=0x0002	## r01
export mono=0x0008	## r01

export tune_on=0x0200	## r02
export tune_off=0x0000	## r02

export R03=0x8012
export R14=0x0c2d	## r14

export R11=0x4f50
export loside=0x0005				## r11
export hiside=0x8000				## r11
export sample_loside=0				## r11
export sample_hiside=$((loside|hiside))		## r11

function init_radio(){
	$FMRADIO -w r01 $((R01 | muted))	\
		 -w r02 0xF6B9			\
		 -w r03 $R03			\
		 -w r04 0x0400			\
		 -w r05 0x28aa			\
		 -w r06 0x4400			\
		 -w r07 0x1ee7			\
		 -w r08 0x7141			\
		 -w r09 0x007d			\
		 -w r10 0x82ce			\
		 -w r11 $((R11 | loside))	\
		 -w r12 0x9700			\
		 -w r13 0xb845			\
		 -w r14 $((R14 | 0xf000))	\
		 -w r15 0x8097			\
		 -w r16 0x04a1			\
		 -w r17 0xdf6a			\
		 -w r00 0xffff			\
	&& wait_for_tune_complete
}

function translate_callsign() {
	freq="$1"
	freq=`echo "$freq" | awk '{print toupper($1)}'`
	orig="$freq"
	[ "$freq" = "LIVE"   ] && freq=885
	[ "$freq" = "CHUO"   ] && freq=891
	[ "$freq" = "HOT"    ] && freq=899
	[ "$freq" = "CBOF"   ] && freq=907
	[ "$freq" = "CBC1"   ] && freq=915
	[ "$freq" = "RADIO1" ] && freq=915
	[ "$freq" = "JACK"   ] && freq=923
	[ "$freq" = "CKCU"   ] && freq=931
	[ "$freq" = "BOB"    ] && freq=939
	[ "$freq" = "CIMF"   ] && freq=949
	[ "$freq" = "JEWEL"  ] && freq=985
	[ "$freq" = "CHRI"   ] && freq=991
	[ "$freq" = "DAWG"   ] && freq=1019
	[ "$freq" = "MAJIC"  ] && freq=1003
	[ "$freq" = "Y101"   ] && freq=1011
	[ "$freq" = "CBC2"   ] && freq=1033
	[ "$freq" = "RADIO2" ] && freq=1033
	[ "$freq" = "KISS"   ] && freq=1053
	[ "$freq" = "CHEZ"   ] && freq=1061
	[ "$freq" = "VIRGIN" ] && freq=1069
	[ "$freq" = "CKDJ"   ] && freq=1079
	if [ "$freq" = "$orig" -o "$freq" = "" ]; then
		nonums=`echo "$freq" | sed '-es/[^0-9]//g'`
		if [ "$freq" != "$nonums" ]; then
			echo "$freq: unknown station callsign" >&2
			exit 1
		fi
		if [ $freq -lt 875 -o $freq -gt 1080 ]; then
			echo "$freq: frequency not in FM band" >&2
			exit 1
		fi
	fi
	echo "$freq"
}

export FMRADIO=fmradio_bitbang
while [ "$1" = "-v" ]; do
	FMRADIO="$FMRADIO -v"
	shift
done

init=
if [ "$1" = "--init" ]; then
	init=yes
	shift
fi

## wait for the radio to power-on:
timeout=6  ## half-seconds
while [ $timeout -gt 0 ]; do
	devid="`$FMRADIO -r devid`"
	[ "$devid" = "0x65b1" ] && break
	init=yes
	sleep 0.5
	timeout=$((timeout - 1))
done
if [ "$devid" != "0x65b1" ]; then
	echo "Radio chip not found"
	exit 1
fi

if [ "$1" = "--status" ]; then
	shift
	report_status
	exit
fi

# ## ensure it's the correct version of the chip:
# chipid="`$FMRADIO -r chipid`"
# if [ "$chipid" != "0x1010" ]; then
# 	echo "Radio chip not found"
# 	exit 1
# fi

## ensure it's not in standby:
[ "$init" != "yes" ] && [ $((`$FMRADIO -r r00` & 1)) -eq 0 ] && init=yes

if [ "$init" != "yes" ]; then
	status="`$FMRADIO -r status`"
	[ "$status" = "0x0000" -o "$status" = "0xffff" ] && init=yes
fi
[ "$init" = "yes" ] && init_radio

if [ "$1" = "--off" ]; then
	$FMRADIO -w r00 0xfffe
	exit 0
fi

if [ "$1" = "--vol" ]; then
	shift
	step=$1
	if [ "$step" != "" ]; then
		shift
		set_volume $step
	fi
	[ "$1" = "" ] && exit
fi

freq="$1"
[ "$freq" = "" ] && freq="CBC1"

freq=`echo "$freq" | sed '-es/[^a-zA-Z0-9]//g'`
freq=`translate_callsign "$freq"`
[ "$freq" != "" ] && autotune "$freq"
