#!/bin/bash

FMRADIO=/usr/local/bin/fmradio_control.sh
AMIXER=/usr/bin/amixer
ARECORD=/usr/bin/arecord
LAME=/usr/bin/lame
LOGGER=/usr/bin/logger

## This program adds extra buffering into the recording/transcoding pipeline:
ELASTICBUFFER=/usr/local/bin/elasticbuffer
[ ! -x $ELASTICBUFFER ] && ELASTICBUFFER=cat

function prefetch_code(){
	cat $ARECORD $ELASTICBUFFER $LAME > /dev/null &
	[ $debug -ne 0 ] && cat $AMIXER > /dev/null &
}

function wait_until_startsecs(){
	secs=$1
	prefetched=
	now=`date '+%s'`
	target=$(((now - (now % 60)) + secs))  ## one second early, plus one leap second.. works well.
	while [ $now -lt $target ]; do
		if [ $((target - now)) -gt 5 ]; then
			prefetch_code
			sleep $((target - now - 3))
		elif [ $((target - now)) -gt 2 ]; then
			prefetch_code
			sleep $((target - now - 1))
		elif [ "$prefetched" = "" ]; then
			prefetch_code
			prefetched=done
		else
			sleep 0.2
		fi
		now=`date '+%s'`
	done
}

function prepare_mixer(){
	## Kill off the sound server and desktop mixer; then reinit the mixer to a known/good state:
	killall artsd &>/dev/null && sleep 0.5
	killall kmix  &>/dev/null && sleep 0.5

	## Configure sound system (slow) to a known-good state:
	#/usr/local/bin/amixer_restore_setup.sh

	#[ $debug -ne 0 ] && $AMIXER cset name='Headphone Playback Switch' 'off'

	## These three settings are the *only* ones that matter when recording from line-in:
	$AMIXER cset name='Capture Source' '4,4'  ## (1=cd,2=video,3=aux,4=line,5=mixer,6=mixmono,7=phone)
	$AMIXER cset name='Capture Volume' '80%'  ## 12,12, max=15,15
	$AMIXER cset name='Capture Switch' 'on'   ## on/off

	## These are optional, needed only if we want to monitor (listen) while recording:
	$AMIXER cset name='Headphone Playback Switch' 'on'
	$AMIXER cset name='Headphone Playback Volume' '96%'
	$AMIXER cset name='Line Playback Switch' 'on'
	$AMIXER cset name='Line Playback Volume' '100%'
}

function do_recording(){
	## Prepare the "sound card" for recording from the line-in jack:
	prepare_mixer &
	prefetch_code &
	mixer=$!

	## Initialize the FM Radio device, and tune to the station (3 attempts max):
	$FMRADIO --init --vol 16 $station || $FMRADIO --init $station || $FMRADIO --init $station || exit 1

	## Select encoding parms differently for stereo vs. mono stations:
	status="`$FMRADIO --status`"
	if [ "${status:0:6}" = "stereo" ]; then
		channels='-c2'
	else
		channels='-c1'
	fi

	## L.A.M.E. recommended settings for FM Radio (from http://jthz.com/mp3/#FM):
	## quality='-V5 --vbr-new -q0 --lowpass 15.4 -b96'	## large  size output
	## quality='-V6 --vbr-new -q0 --lowpass 15.4 -b96'	## medium size output
	## quality='-V7 --vbr-new -q0 --lowpass 15.4 -b112 '	## small  size output

	  if [ "$quality" = "high" ]; then
		quality='-V4 --vbr-new -q0 --lowpass 15.4 -b96'
	elif [ "$quality" = "medium" ]; then
		quality='-V6 --vbr-new -q0 --lowpass 15.4 -b64'
	elif [ "$quality" = "low" ]; then
		quality='-V8 --vbr-new -q0 --lowpass 14.0 -b56'
	else ## "lowest" or whatever
		quality='-V9 --vbr-new -q0 --lowpass 12.0 -b48'
	fi

	## The "sound card" init should have finished long before now:
	wait $mixer

	## Match the specified "seconds on the clock" for a near-exact start time:
	[ "$startsecs" != "" ] && wait_until_startsecs $startsecs

	$LOGGER -s -- "Recording $station ($status) to \"$out\" for $minutes mins ($seconds secs)." &
	[ $debug -ne 0 ] && $AMIXER cset name='Headphone Playback Switch' 'on'
	$ARECORD --quiet --duration=$seconds -f S16_LE $channels -r44100 \
		| $ELASTICBUFFER | $LAME --silent $quality - $out

	$LOGGER -s -- "Finished recording \"$out\"."
	$FMRADIO --off
}

debug=0
seconds=
quality=medium

while [ "${1:0:1}" = "-" ]; do
	arg="$1"
	shift
	  if [ "$arg" = "--debug" ]; then
		debug=1
	elif [ "$arg" = "--startsecs" -a "$1" != "" ]; then
		startsecs="$1"
		shift
	elif [ "$arg" = "--quality"   -a "$1" != "" ]; then
		quality="$1"
		shift
	else
		echo "$arg: unknown flag, aborting"
		exit 1
	fi
done

if [ "$1" = "" -o "$2" = "" -o "$3" = "" -o "$4" != "" ]; then
	echo "Usage: $0 [--debug] [--startsecs nn] [--quality {0..3}] <station|freq> <minutes> <outprefix>"
	exit 1
fi

station="$1"
minutes="$2"
prefix="$3"
seconds=$((minutes * 60))
out="${prefix}.`date "+%Y-%m-%d_%H%M"`.mp3"
if [ $debug -eq 0 ]; then
	do_recording &>/dev/null </dev/null &
else
	do_recording
fi
exit
